#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 11 17:45:10 2019

@author: felix and jonas
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D #Braucht man für 3D Plots
from scipy.linalg import norm


class Vertex:
    def __init__(self, coords = np.zeros(3), edge = None):
        self.edge = edge
        self.coords = coords
        
    def __str__(self):
        return str(self.coords)
        
    def numofEdges(self):
        he = self.edge
        #hcurr = he
        hnext = he.twinEdge.nextEdge
        count =1
        while not hnext is he:
            count = count +1
            #hcurr = hnext
            hnext = hnext.twinEdge.nextEdge
        return count
        

    
    def isboundary(self):
        he = self.edge
        hcurr = he
        hnext = he.twinEdge.nextEdge
        if he.twinEdge.face is None:
            return True
        
        while not hnext is he:
            hcurr = hnext
            hnext = hnext.twinEdge.nextEdge
            if hcurr.twinEdge.face is None:
               return True
        return False
    
class Face:
    def __init__(self, halfedge = None, normal = None):
        self.edge = halfedge
        self.normal = normal
    
    def __str__(self):
        string = "The face containes the edge: "+ str(self.edge.edgeVector())
        return string
        
    def setnormal(self):
        if self.isplanar():    
            e1 = self.edge.edgeVector()
            e2 = self.edge.nextEdge.edgeVector()
            v = np.cross(e1,e2)
            v = v / norm(v)
            self.normal = v
        else:
            print("the face is not planar")
        
    def numedges(self):
        count =1
        he= self.edge
        hecurr = he.nextEdge
        count =count +1
        while not hecurr.nextEdge is he:
            hecurr = hecurr.nextEdge
            count = count +1
        return count    
        
                
    def isplanar(self):
        b = True
        he = self.edge
        he1 = he
        he2 = he1.nextEdge
        he3 = he2.nextEdge
        a = np.array([he1.edgeVector(), he2.edgeVector(), he3.edgeVector()])
        det = np.linalg.det(a)
        if det is 0:
            b= False
            return b
        while not he3 is he:
            he1 = he2
            he2 = he3
            he3 = he3.nextEdge
            a = np.array([he1.edgeVector(), he2.edgeVector(), he3.edgeVector()])
            det = np.linalg.det(a)
            if det is 0:
                b= False
                return b
        return b  
        
    def getCenterofMass(self):
        n = self.numedges()
        if self.isplanar:
            he=self.edge
            v1=he.vertex.coords
            hcurr=he.nextEdge
            while not hcurr.nextEdge is he:
                v1 = v1+ hcurr.vertex.coords
                hcurr = hcurr.nextEdge
            v1 = (v1 + hcurr.vertex.coords)/n
            return v1
        print("the Face is not planar")
        return None

class Halfedge:
    def __init__(self, nextEdge = None, twinEdge = None, point = None, face = None):
        self.nextEdge = nextEdge #Naechste Kante
        self.twinEdge = twinEdge #Zwillingskante
        self.vertex = point #Der Punkt auf den die Kante zeigt
        self.face = face #Zugehoerige Flaeche
        
    #Give a string that says which two points are connected by the edge
    def __str__(self):
        edges = "The edge points from "+str(self.sourcepoint().coords)+" to "+str(self.vertex.coords)
        return edges

    #Gives the sourcepoint of the halfedge using the twinEdge
    def sourcepoint(self):
        if self.twinEdge is None:
            hcurr = self.nextEdge
            hnext = hcurr.nextEdge
            #Continue until the next edge is the initial edge
            count = 0
            while(not hnext is self and count < 100):
                hcurr = hnext
                hnext = hcurr.nextEdge
                count += 1 #Just for savety reasons...
            #Return the point of the current edge
            return hcurr.vertex
        else:
            return self.twinEdge.vertex

    #Gives the edgevector as an np.array of the halfedge using the twinEdge
    def edgeVector(self):
        v1 = self.vertex.coords
        v2 = self.sourcepoint().coords
        return v1-v2

    #Gives the previous edge of the halfedge by iterating ofer the face and using nextEdge. This works for halfedges without twinEdges
    #and is used in the methode that creates a halfedges mesh from an obj file
    def previousEdge(self):
        hcurr = self.nextEdge
        count = 0
        while(not hcurr.nextEdge is self and count < 100):
            hcurr = hcurr.nextEdge
            count += 1
        return hcurr

    #Gives length of the halfedge
    def length(self):
        l = norm(self.edgeVector())
        return l
    
    
    def angle(self):
        if self.face is None:
            print("boundary edge")
            v1 = self.edgeVector()
            v1 = v1/norm(v1)
            v2 = self.nextEdge.edgeVector()
            v2 = v2/norm(v2)
            a = np.dot(v1,v2)
            alpha = np.arccos(a)+np.pi
        else:
            v1 = self.edgeVector()
            v1 = -v1/norm(v1)
            v2 = self.nextEdge.edgeVector()
            v2 = v2/norm(v2)
            a = np.dot(v1,v2)
            alpha = np.arccos(a)
        return alpha
    
       
            
    def isboundary(self):
        if self.face is None:
           return True
        else:
           return False         
                       
class Mesh:
    def __init__(self):

        self.vertices = []
        self.faces = []
        self.edges = []
            
    def coords(self):
        #Zuerst Liste anlegen:
        coords = []
        for v in self.vertices:
            coords.append(v.coords)
        coords = np.array(coords)
        return coords
    
        
    def centerofFaces(self):
        c = []
        for f in self.faces:
            c.append(f.getCenterofMass())        
        c = np.array(c).astype(float)
        #print(c)
        return c
    
    def edges2(self):
        #Zuerst Liste anlegen:
        edges = "List of edges: "
        i = 0
        for e in self.edges:
            edges = edges +str(i)+" : from "+str(e.sourcepoint().coords)+" to "+str(e.vertex.coords)+", "
            i = i+1 
        
        return edges                
 
#---------------------------------------------------------------------------------------------------------------------                                                 
                                            
    def plot(self, color = 'b', points = True, center = True, normals = True, pointnum = True, edgenum = True):
        figure = plt.figure()
        ax = figure.add_subplot(111, projection='3d')
        ax.set_aspect(1)
        
        if points:
            ax.scatter3D(self.coords()[:,0], self.coords()[:,1], self.coords()[:,2], color = 'r')
            
        if pointnum:
            n = 0
            for v in self.vertices:
                ax.text(self.coords()[n,0]+0.05, self.coords()[n,1],self.coords()[n,2],n, fontsize=9)
                n = n+1
    
        for face in self.faces:
            edge = face.edge
            #Die Punkte fuer die erste Kante waehlen:
            v0 = edge.vertex 
            v1 = edge.nextEdge.vertex
            #Die erste Kante plotten:
            ax.plot3D([v0.coords[0],v1.coords[0]],[v0.coords[1], v1.coords[1]],[v0.coords[2], v1.coords[2]], color = color)
            
            #Eine Kante weiter gehen:
            edge = edge.nextEdge
            v0 = v1
            v1 = edge.nextEdge.vertex
            while (not v0 is face.edge.vertex): #Ueberpruefen, ob wir bei der Anfangskante angelangt sind.
                #Kante plotten:
                ax.plot3D([v0.coords[0],v1.coords[0]],[v0.coords[1], v1.coords[1]],[v0.coords[2], v1.coords[2]], color = color)
                #Eine Kante weiter gehen:
                edge = edge.nextEdge
                v0 = v1
                v1 = edge.vertex
                
        if center:
            c = self.centerofFaces()
            ax.scatter3D(c[:,0], c[:,1], c[:,2], color = 'g')
                
        if normals:
                
            c = self.centerofFaces()
            self.setnormals()
            N = self.getnormals()
            n = self.numfaces()
            for i in range(n):  
                ax.quiver(c[i,0], c[i,1], c[i,2], N[i,0], N[i,1], N[i,2],length=1, normalize=True)
                
        '''#Alternative kann man die Normalen auch als Linien plotten...
                c = self.centerofFaces()
                self.setnormals()
                N = self.getnormals()
                n = self.numfaces()
                for i in range(n):
                    ax.plot3D([c[i,0],c[i,0]+N[i,0]],[c[i,1],c[i,1]+N[i,1]],[c[i,2],c[i,2]+N[i,2]], color = 'r')
        '''
        ax.set_xlabel('X axis')
        ax.set_ylabel('Y axis')
        ax.set_zlabel('Z axis')
        
        plt.show()
        return figure
    
    
    #Small Functions:-------------------------------------------------------------------------------------
    
    #Gives the number of faces of the mesh
    def numfaces(self): return len(self.faces)
    
    #Gives the number of vertices of the mesh
    def numvertices(self): return len(self.vertices)
        
    #Gives the number of halfedges of the mesh
    def numedges(self): return len(self.edges)
    
    #Return the number of a vertex, face or edge:
    def vertexnum(self, vertex):
        return self.vertices.index(vertex)
        
    def edgenum(self, edge):
        return self.edges.index(edge)
        
    def facenum(self, face):
        return self.faces.index(face)
    

    
    def setnormals(self):
        for f in self.faces:
            f.setnormal()
        
            
    def getnormals(self):
         N = []
         for f in self.faces:
             N.append(f.normal)
         N = np.array(N).astype(float)
         #print(str(N))
         return N
       

             
#-----------------------------------------------------------------        
    
    #Eine Obj-Datei einlesen:
    def halfedgeFromObj(self, Dateipfad):
        #Initialzierung der Obj-Listen
        V = []
        F = []
        #Initialisierung der halfedge-Listen
        vertices = []
        faces = []
        halfedges = []
        #Auslesen des obj files
        file = open( Dateipfad, 'r')
        for line in file:
            if line.startswith('v '):
                V.append(line.split()[1:])
            if line.startswith('f '):
                f = line.split()[1:]
                for i in range(len(f)): 
                    f[i] = f[i].split('/')[0] #Only use the vertex coordinates of f 1/2/3 4/5/6 17/8/9
                F.append(f)
        V = np.array(V).astype(float)
        #F = np.array(F).astype(int)
        #print(F)
        #print(V)
        k = len(V)

        connect = np.ones( (k,k) )
        connect = -1*connect

        #safe the points as vertices in the vertices list
        for v in V:
            ver = Vertex(v,None)
            vertices.append(ver)

        #iterate over F to create and save the faces and halfedges. at first the halfedges do not have an opposite edge
        for f in F:
            #write an exception if the face is not a triangle
            #if len(f) != 3:
            #    print("this is not a triangle")
            n = len(f) #Number of vertices in face f
            
            #get the current position in the list of faces and halfedges
            lh = len(halfedges)
            lf = len(faces)

            #create the new face and add it to the list
            face = Face()
            faces.append(face)

            #create new halfedges, without opposite edge and next edge:
            for i in range(n):
                halfedges.append(Halfedge(None,None,vertices[int(f[i])-1],faces[lf])) # -1 because obj files start counting by 1
            
            #Connect the inner halfedges:
            for i in range(n):
                halfedges[lh +i].nextEdge = halfedges[lh + (i+1)%n]

            #Assign the out going edge to the points:
            for i in range(n):
                vertices[int(f[i])-1].edge = halfedges[lh + (i+1)%n]

            #Assign one of the halfedges to the face
            faces[lf].edge = halfedges[lh]

            # if the halfedge points from the vertex i to the vertex j we save it in the connect matrix at the point (i,j)
            for i in range(n):
                connect[int(f[(i-1)%n])-1, int(f[i])-1] = lh + i

        #Now we set the oposit edges for the interior edges
        for i in range(k):
            for j in range(i,k):
                if connect[i,j] > -1 and connect[j,i] > -1 :
                        halfedges[int(connect[i,j])].twinEdge = halfedges[int(connect[j,i])]
                        halfedges[int(connect[j,i])].twinEdge = halfedges[int(connect[i,j])]

        #Now we create the oposit edges for the exterior edges and add them to the list
        for h in halfedges:
            #first check if h is an exteriour edge,i.e. h has no twin
            if  h.twinEdge is None :
                #create a new halfedge that points on the source of h. The new edge does not have a face because it is an outer boundary edge
                #We cannot assign the nextEdge yet because it does not exists so far.
                halfedges.append(Halfedge(None,h,h.sourcepoint(),None))

                #save the position of the new halfedge such that we can later define its next edge and define the counter curr
                pos = len(halfedges)-1
                curr = len(halfedges)-1

                #assign the new edge to h as its twinEdge
                h.twinEdge = halfedges[pos]
                #Now we walk along the boundary and create the missing edges. Therefore we consider the previous edge of h and check if its part of the boundary.
                hp = h.previousEdge()
                while( not hp is h):
                    #if hp is part of the boundary we create its twin edge
                    if hp.twinEdge is None :
                        halfedges.append(Halfedge(None,hp,hp.sourcepoint(),None))
                        halfedges[curr].nextEdge = halfedges[curr+1]
                        curr = curr+1
                        hp.twinEdge = halfedges[curr]
                        hp = hp.previousEdge()
                    #if hp is not part of the boundary we continue with next outgoing edge in the vertexstar
                    else:
                        hp = hp.twinEdge.previousEdge()
                #know we can close the circle and give the boundary edge that we created first its nextEdge
                if hp is h:
                    halfedges[curr].nextEdge = halfedges[pos]

        #Finally we can assigne the new lists to the mesh
        self.vertices = vertices
        self.faces = faces
        self.edges = halfedges
        return self
        
        
        
#----------------------------------------------------------------------
# Ein Mesh als Obj Datei exportieren

    def exportobj(self, Dateipfad):
        
        file = open(Dateipfad,"w")
        v= len(self.vertices) 
        f= len(self.faces) 
        h= len(self.edges)/2
        file.write("# Number of vertices: " + str(v) +"\n" )
        file.write("# Number of faces: " + str(f) +"\n" ) 
        file.write("# Number of edges: " + str(h) +"\n" )        
        
        
        for p in self.vertices:
            file.write("v " + str(p.coords[0]) + " " + str(p.coords[1]) + " " + str(p.coords[2]) +"\n")
   
        
        file.write("\n"+ "# Das waren die Ecken jetzt kommen die Facetten" + "\n\n")
        for f in self.faces:
            file.write("f ")
            he = f.edge
            v = self.vertexnum(he.vertex) +1
            file.write(str(v))
            hcurr = he.nextEdge
            while hcurr != he:
                v = self.vertexnum(hcurr.vertex) +1
                file.write(" "+str(v))
                hcurr = hcurr.nextEdge  
                
            file.write("\n")
        file.close()
 
#-----------------------------------------------------------------------------------------------
#Funktions to transform the mesh
   
   
    def scale(self,s=2):
         for v in self.vertices:
             v.coords = s*v.coords

    def translate(self, vektor = np.array([1,1,1])):
        for v in self.vertices:
             v.coords = vektor + v.coords
             
    def inverse(self):
        for v in self.vertices:
             if norm( v.coords) is 0 :
                 i=v.vertexnum()
                 print("Der Punkt " + str(i) + " wird nach unendlich abgebildet")
             else:
                v.coords = v.coords/ pow(norm( v.coords),2)
 
 
 
 
 
 
class Quad(Mesh):
    
    
    
    def __init__(self, points = [[0,0,0],[1,0,0],[1,1,0],[0,1,0]]):
        Mesh.__init__(self) #Ruft den Mesh Initialisator auf.
        
        if type(points) == list:
            points = np.array(points)
        
        self.faces.append(Face())
        
        #Punkte anlegen
        for i in range(4):
            self.vertices.append(Vertex(points[i]))
        
        #Innen-Kanten anlegen:
        self.edges.append(Halfedge(None,None,self.vertices[0],self.faces[0]))
        self.edges.append(Halfedge(self.edges[0],None,self.vertices[3],self.faces[0]))
        self.edges.append(Halfedge(self.edges[1],None,self.vertices[2],self.faces[0]))
        self.edges.append(Halfedge(self.edges[2],None,self.vertices[1],self.faces[0]))
        self.edges[0].nextEdge = self.edges[3]
        
        #Jedem Punkt die ausgehende Innen-Kante zuordnen:
        self.vertices[0].edge = self.edges[3]
        self.vertices[1].edge = self.edges[2]
        self.vertices[2].edge = self.edges[1]            
        self.vertices[3].edge = self.edges[0]
        
        #Außenkanten setzen:
        for i in range(4):
            self.edges.append(Halfedge(None,self.edges[i],self.vertices[3-i],None))
            
        #Aussenkanten verknuepfen:
        for i in range(4):
            self.edges[i+4].nextEdge = self.edges[(i+1)%4+4] #Wir rechnen modulo n um im Kreis zu gehen.
                
        #Zwillingskanten zuordnen:
        for i in range(4):
            self.edges[i].twinEdge = self.edges[i + 4]
        
        #Dem face eine Halfedge zuordnen:
        self.faces[0].edge = self.edges[0]

        
class Triangle(Mesh):
    
    def __init__(self, points = [[0,0,0],[1,0,0],[0,1,0]]):
        Mesh.__init__(self) #Ruft den Mesh Initialisator auf.
        
        self.faces.append(Face())
        
        if type(points) == list:
            points = np.array(points)
        
        #Punkte anlegen
        for i in range(3):
            self.vertices.append(Vertex(points[i]))
            
        #Innenkanten anlegen:
        self.edges.append(Halfedge(None,None,self.vertices[0],self.faces[0])) #Kante 0
        self.edges.append(Halfedge(self.edges[0],None,self.vertices[2],self.faces[0])) #Kante 1
        self.edges.append(Halfedge(self.edges[1],None,self.vertices[1],self.faces[0])) #Kante 2
        self.edges[0].nextEdge = self.edges[2]
        
        #Jedem Punkt die ausgehende Innenkante zuordnen:
        for i in range(3):
            self.vertices[i].edge = self.edges[2-i]
        
        #Aussenkanten setzen:
        for i in range(3):
            self.edges.append(Halfedge(None,self.edges[i],self.vertices[2-i],None))
            
        #Aussenkanten verknuepfen:
        for i in range(3):
            self.edges[i+3].nextEdge = self.edges[(i+1)%3+3] #Wir rechnen modulo n um im Kreis zu gehen.
        
        
        #Zwillingskanten zuordnen:       
        for i in range(3):
            self.edges[i].twinEdge = self.edges[i+3]
        
        #Dem face eine Halfedge zuordnen:
        self.faces[0].edge = self.edges[0]
        
        
           

        
class Ngon(Mesh):
    
    def __init__(self, points = np.array([[0,0,0],[1,0,0],[0,1,0]])):
        '''
        Die Punkte werden gegen den Uhrzeigersinn durch nummeriert. Die Innenkanten ebenfalls,
        wobei die i-te Kante auf den i-ten Punkt zeigt.
        Die Aussenkanten laufen auch gegen den Uhrzeigersinn, wobei die (n+i)-te Kante Zwilling zur i-ten
        Kante ist.
        '''
        if type(points) == list:
            points = np.array(points)
        
        Mesh.__init__(self) #Ruft den Mesh Initialisator auf.
        n = points.shape[0]
        
        self.faces.append(Face())
        
        #Punkte anlegen
        for i in range(n):
            self.vertices.append(Vertex(points[i]))
        
        #Innenkanten anlegen:
        for i in range(n):
            self.edges.append(Halfedge(None,None,self.vertices[i],self.faces[0])) #Kante i zeigt auf Punkt i
        
        #Innenkanten verknuepfen:
        for i in range(n):
            self.edges[i].nextEdge = self.edges[(i+1)%n] #Wir rechnen modulo n um im Kreis zu gehen.
        
        #Aussenkanten anlegen:
        for i in range(n):
            self.edges.append(Halfedge(None,self.edges[i],self.vertices[(n-1+i)%n],None))
        
        #Aussenkanten verknuepfen:
        for i in range(n):
            self.edges[i+n].nextEdge = self.edges[(i+1)%n+n] #Wir rechnen modulo n um im Kreis zu gehen.
        
        #Jedem Punkt eine ausgehende Innenkante zuordnen:
        for i in range(n):
            self.vertices[i].edge = self.edges[(i+1)%n]
     
        #Den Innenkanten die Zwillingskanten zuordnen:        
        for i in range(n):
            self.edges[i].twinEdge = self.edges[i+n]
            
       
                
        #Dem face einen Punkt zuordnen:
        self.faces[0].edge = self.edges[0]

############################## Test Area ###########################################

mesh1 = Mesh().halfedgeFromObj(r"C:/Users/jtervooren/Desktop/python/obj_files/torus1.obj")

mesh1.scale(1)
mesh1.translate([1,1,0])
mesh1.inverse()



mesh1.plot(points = False, pointnum = False, normals = False, center = False)
#mesh.setnormals()


            
    
#mesh1.exportobj(r"C:/Users/jtervooren/Desktop/python/obj_files/test.obj")